import io
import unittest

from haoda.report.xilinx import rtl


class TestParser(unittest.TestCase):

  def setUp(self):
    pass

  def test_dir_rpt(self):
    dir_rpt_file = io.StringIO(
        """Copyright 1986-2020 Xilinx, Inc. All Rights Reserved.
------------------------------------------------------------------------------------
| Tool Version : Vivado v.2020.2 (lin64) Build 3064766 Wed Nov 18 09:12:47 MST 2020
| Date         : Thu Jun 10 11:51:34 2021
| Host         : haodahost running 64-bit Ubuntu 18.04.5 LTS
| Command      : report_utilization -file report.rpt -hierarchical
| Design       : Top
| Device       : xcu280fsvh2892-2L
| Design State : Optimized
------------------------------------------------------------------------------------

Utilization Design Information

Table of Contents
-----------------
1. Utilization by Hierarchy

1. Utilization by Hierarchy
---------------------------

+-----------------------------------+------------------------------+------------+------------+---------+------+------+--------+--------+------+------------+
|              Instance             |            Module            | Total LUTs | Logic LUTs | LUTRAMs | SRLs |  FFs | RAMB36 | RAMB18 | URAM | DSP Blocks |
+-----------------------------------+------------------------------+------------+------------+---------+------+------+--------+--------+------+------------+
| Top                               |                        (top) |       8354 |       6026 |    2328 |    0 | 5005 |      2 |      1 |    4 |          3 |
|   (Top)                           |                        (top) |       1097 |       1097 |       0 |    0 | 4423 |      0 |      1 |    0 |          1 |
|   SubModule1                      |               Top_SubModule1 |       7176 |       4848 |    2328 |    0 |  582 |      2 |      0 |    0 |          1 |
|     SubSubModule1                 |            Top_SubSubModule1 |       7176 |       4848 |    2328 |    0 |  582 |      2 |      0 |    0 |          1 |
|   SubModule2                      |               Top_SubModule2 |         36 |         36 |       0 |    0 |    0 |      0 |      0 |    2 |          1 |
|   SubModule3                      |               Top_SubModule3 |         45 |         45 |       0 |    0 |    0 |      0 |      0 |    2 |          0 |
+-----------------------------------+------------------------------+------------+------------+---------+------+------+--------+--------+------+------------+
""")
    top_total = rtl.parse_hierarchical_utilization_report(dir_rpt_file)
    schema = top_total.schema

    self.assertDictEqual(
        schema,
        {
            'Module': 0,
            'Total LUTs': 1,
            'Logic LUTs': 2,
            'LUTRAMs': 3,
            'SRLs': 4,
            'FFs': 5,
            'RAMB36': 6,
            'RAMB18': 7,
            'URAM': 8,
            'DSP Blocks': 9,
        },
    )

    self.assertIsNone(top_total.parent)
    self.assertEqual(len(top_total.children), 4)
    self.assertEqual(top_total.instance, 'Top')
    self.assertEqual(top_total['Module'], '(top)')
    self.assertEqual(top_total['Total LUTs'], '8354')
    self.assertEqual(top_total['Logic LUTs'], '6026')
    self.assertEqual(top_total['LUTRAMs'], '2328')
    self.assertEqual(top_total['SRLs'], '0')
    self.assertEqual(top_total['FFs'], '5005')
    self.assertEqual(top_total['RAMB36'], '2')
    self.assertEqual(top_total['RAMB18'], '1')
    self.assertEqual(top_total['URAM'], '4')
    self.assertEqual(top_total['DSP Blocks'], '3')

    top_self, sub_module_1, sub_module_2, sub_module_3 = top_total.children

    self.assertIs(top_self.parent, top_total)
    self.assertListEqual(top_self.children, [])
    self.assertEqual(top_self.instance, '(Top)')
    self.assertIs(top_self.schema, schema)
    self.assertEqual(top_self['Module'], '(top)')
    self.assertEqual(top_self['Total LUTs'], '1097')
    self.assertEqual(top_self['Logic LUTs'], '1097')
    self.assertEqual(top_self['LUTRAMs'], '0')
    self.assertEqual(top_self['SRLs'], '0')
    self.assertEqual(top_self['FFs'], '4423')
    self.assertEqual(top_self['RAMB36'], '0')
    self.assertEqual(top_self['RAMB18'], '1')
    self.assertEqual(top_self['URAM'], '0')
    self.assertEqual(top_self['DSP Blocks'], '1')

    self.assertIs(sub_module_1.parent, top_total)
    self.assertEqual(len(sub_module_1.children), 1)
    self.assertEqual(sub_module_1.instance, 'SubModule1')
    self.assertIs(sub_module_1.schema, schema)
    self.assertEqual(sub_module_1['Module'], 'Top_SubModule1')
    self.assertEqual(sub_module_1['Total LUTs'], '7176')
    self.assertEqual(sub_module_1['Logic LUTs'], '4848')
    self.assertEqual(sub_module_1['LUTRAMs'], '2328')
    self.assertEqual(sub_module_1['SRLs'], '0')
    self.assertEqual(sub_module_1['FFs'], '582')
    self.assertEqual(sub_module_1['RAMB36'], '2')
    self.assertEqual(sub_module_1['RAMB18'], '0')
    self.assertEqual(sub_module_1['URAM'], '0')
    self.assertEqual(sub_module_1['DSP Blocks'], '1')

    sub_sub_module_1 = sub_module_1.children[0]

    self.assertIs(sub_sub_module_1.parent, sub_module_1)
    self.assertListEqual(sub_sub_module_1.children, [])
    self.assertEqual(sub_sub_module_1.instance, 'SubSubModule1')
    self.assertIs(sub_sub_module_1.schema, schema)
    self.assertEqual(sub_sub_module_1['Module'], 'Top_SubSubModule1')
    self.assertEqual(sub_sub_module_1['Total LUTs'], '7176')
    self.assertEqual(sub_sub_module_1['Logic LUTs'], '4848')
    self.assertEqual(sub_sub_module_1['LUTRAMs'], '2328')
    self.assertEqual(sub_sub_module_1['SRLs'], '0')
    self.assertEqual(sub_sub_module_1['FFs'], '582')
    self.assertEqual(sub_sub_module_1['RAMB36'], '2')
    self.assertEqual(sub_sub_module_1['RAMB18'], '0')
    self.assertEqual(sub_sub_module_1['URAM'], '0')
    self.assertEqual(sub_sub_module_1['DSP Blocks'], '1')

    self.assertIs(sub_module_2.parent, top_total)
    self.assertListEqual(sub_module_2.children, [])
    self.assertEqual(sub_module_2.instance, 'SubModule2')
    self.assertIs(sub_module_2.schema, schema)
    self.assertEqual(sub_module_2['Module'], 'Top_SubModule2')
    self.assertEqual(sub_module_2['Total LUTs'], '36')
    self.assertEqual(sub_module_2['Logic LUTs'], '36')
    self.assertEqual(sub_module_2['LUTRAMs'], '0')
    self.assertEqual(sub_module_2['SRLs'], '0')
    self.assertEqual(sub_module_2['FFs'], '0')
    self.assertEqual(sub_module_2['RAMB36'], '0')
    self.assertEqual(sub_module_2['RAMB18'], '0')
    self.assertEqual(sub_module_2['URAM'], '2')
    self.assertEqual(sub_module_2['DSP Blocks'], '1')

    self.assertIs(sub_module_3.parent, top_total)
    self.assertListEqual(sub_module_3.children, [])
    self.assertEqual(sub_module_3.instance, 'SubModule3')
    self.assertIs(sub_module_3.schema, schema)
    self.assertEqual(sub_module_3['Module'], 'Top_SubModule3')
    self.assertEqual(sub_module_3['Total LUTs'], '45')
    self.assertEqual(sub_module_3['Logic LUTs'], '45')
    self.assertEqual(sub_module_3['LUTRAMs'], '0')
    self.assertEqual(sub_module_3['SRLs'], '0')
    self.assertEqual(sub_module_3['FFs'], '0')
    self.assertEqual(sub_module_3['RAMB36'], '0')
    self.assertEqual(sub_module_3['RAMB18'], '0')
    self.assertEqual(sub_module_3['URAM'], '2')
    self.assertEqual(sub_module_3['DSP Blocks'], '0')



import os
import re
import shutil
import tempfile
import xml.etree.ElementTree as ET
import zipfile
from typing import BinaryIO, Dict, Optional, TextIO

from haoda import util
from haoda.backend import xilinx as backend

__all__ = (
    'ReportDirUtil',
    'ReportXoUtil',
    'RtlHlsInfo',
)

REPORT_UTIL_COMMANDS = r'''
read_verilog [ glob {hdl_dir}/*.v ]
set ips [ glob -nocomplain {hdl_dir}/*/*.xci ]
if {{ $ips ne "" }} {{
  read_ip $ips
  generate_target synthesis [ get_files *.xci ]
}}
foreach tcl_file [glob -nocomplain {hdl_dir}/*.tcl] {{
  source $tcl_file
}}
{set_parallel}
synth_design {synth_args}
opt_design
report_utilization {report_util_args}
'''


class ReportDirUtil(backend.Vivado):
  """Run synthesis and generate resource utilization report.

  This class is a child of subprocess.Popen and will launch Vivado to run
  synthesis and generate resource utilization report. Arguments passed to
  synth_design and report_utilization can be configured via the kwargs params.

  Attributes:
    job_server_fd: Optional file descriptor of a job server.
    num_jobs: Optional number of jobs.
  """

  def __init__(
      self,
      hdl_dir: str,
      rpt_path: str,
      top_name: str,
      part_num: Optional[str] = None,
      synth_kwargs: Optional[Dict[str, str]] = None,
      report_util_kwargs: Optional[Dict[str, str]] = None,
  ):
    """Run synthesis and generate resource utilization report.

    Args:
      hdl_dir: HDL directory containing *.v and *.tcl files.
      rpt_file: Path of generated resource utilization report.
      top_name: Top module name.
      part_num: Part number; infered from HDL if empty.
      synth_kwargs: Additional arguments for synth_design.
      report_util_kwargs: Additional arguments for report_utilization.

    Raises:
      InputError if part number cannot be inferred.
    """
    if part_num is None:
      for hdl_file_format in ('{}.v', '{0}_{0}.v'):
        try:
          with open(os.path.join(
              hdl_dir,
              hdl_file_format.format(top_name),
          )) as hdl_file:
            part_num = RtlHlsInfo(hdl_file)['HLS_INPUT_PART']
          break
        except FileNotFoundError:
          pass
    if part_num is None:
      raise util.InputError('failed to infer part number')

    if synth_kwargs is None:
      synth_kwargs = {}
    synth_kwargs['top'] = top_name
    synth_kwargs['part'] = part_num

    if report_util_kwargs is None:
      report_util_kwargs = {}
    report_util_kwargs['file'] = os.path.abspath(rpt_path)
    report_util_kwargs.setdefault('hierarchical', '')

    synth_args = ' '.join(f'-{k} {v}' for k, v in synth_kwargs.items())
    report_util_args = ' '.join(
        f'-{k} {v}' for k, v in report_util_kwargs.items())
    kwargs = {
        'hdl_dir': os.path.abspath(hdl_dir),
        'synth_args': synth_args,
        'report_util_args': report_util_args,
        'set_parallel': '',
    }
    self.job_server_fd = util.get_job_server_fd(())
    if self.job_server_fd is not None:
      new_fd = os.open('/proc/self/fd/%d' % self.job_server_fd,
                       os.O_RDWR | os.O_NONBLOCK)
      self.num_jobs = 1  # maximum supported by Vivado is 8
      try:
        for i in range(7):
          self.num_jobs += len(os.read(new_fd, 1))
      except BlockingIOError as e:
        pass
      os.close(new_fd)
      kwargs['set_parallel'] = 'set_param general.maxThreads %d' % self.num_jobs
    super().__init__(REPORT_UTIL_COMMANDS.format(**kwargs))

  def __exit__(self, *args):
    super().__exit__(*args)
    if self.job_server_fd is not None:
      os.write(self.job_server_fd, b'x' * (self.num_jobs - 1))


class ReportXoUtil(ReportDirUtil):
  """Run synthesis and generate resource utilization report.

  This class is a child of subprocess.Popen and will launch Vivado to run
  synthesis and generate resource utilization report. Arguments passed to
  synth_design and report_utilization can be configured via the kwargs params.

  Attributes:
    tmpdir: Temporary working directory for the RTL files and generated report.
    rpt_file: File object of generated resource utilization report.
    rpt_file_name: Name of the generated temporary resource utilization report.
  """

  def __init__(self,
               xo_file: BinaryIO,
               rpt_file: TextIO,
               top_name: str = 'Dataflow',
               part_num: Optional[str] = None,
               synth_kwargs: Optional[Dict[str, str]] = None,
               report_util_kwargs: Optional[Dict[str, str]] = None):
    """Run synthesis and generate resource utilization report.

    Args:
      xo_file: XO file object containing the HDL files.
      rpt_file: File object of generated resource utilization report.
      top_name: Optionally specify a different top name other than "Dataflow";
          inferred from XO file if empty.
      part_num: Optionally specify a different part number other than
          automatically determined.
      synth_kwargs: Dict of arguments for the synth_design command.
      report_util_kwargs: Dict of arguments for the synth_design command.

    Raises:
      InputError: if input is not a valid XO.
      InternalError: if the report is not generated as expected.
    """
    self.tmpdir = tempfile.TemporaryDirectory(prefix='report-xo-util-')
    self.rpt_file = rpt_file
    self.rpt_file_name = os.path.join(self.tmpdir.name, 'post_synth_util.rpt')
    with zipfile.ZipFile(xo_file) as xo_zip:
      with xo_zip.open('xo.xml') as xo_xml:
        kernel = ET.parse(xo_xml).find('./Kernels/Kernel')
        if kernel is None:
          raise util.InputError('cannot parse XO file')
        ip_dir = kernel.attrib['IP']
        kernel_name = kernel.attrib['Name']
      hdl_dir_prefix = ip_dir + '/src'
      if all(not x.startswith(hdl_dir_prefix) for x in xo_zip.namelist()):
        hdl_dir_prefix = ip_dir + '/hdl/verilog'
      hdl_dir = os.path.join(self.tmpdir.name, hdl_dir_prefix)
      xo_zip.extractall(path=self.tmpdir.name,
                        members=[
                            name for name in xo_zip.namelist()
                            if name.startswith(hdl_dir_prefix)
                        ])

    if not top_name:
      top_name = kernel_name

    super().__init__(
        hdl_dir,
        self.rpt_file_name,
        top_name,
        part_num,
        synth_kwargs,
        report_util_kwargs,
    )

  def __exit__(self, *args):
    super().__exit__(*args)
    try:
      with open(self.rpt_file_name) as src_rpt_file:
        shutil.copyfileobj(src_rpt_file, self.rpt_file)
    except FileNotFoundError:
      raise util.InternalError('failed to generate report file')
    self.tmpdir.cleanup()


RTL_HLS_INFO_REGEX = r'\(\* CORE_GENERATION_INFO\s*=\s*".*,\{(.*)\}" \*\)'


class RtlHlsInfo:

  def __init__(self, rtl_file: TextIO):
    match = re.search(RTL_HLS_INFO_REGEX, rtl_file.read())
    if match is None:
      raise util.InputError('cannot parse RTL file')
    for item in match.group(1).split(','):
      key, value = item.split('=')
      setattr(self, key, value)

  def __getitem__(self, key: str) -> str:
    return getattr(self, key)



import collections

import sys
import enum
from typing import Dict, NamedTuple, Optional, TextIO, List, Tuple
import logging

from haoda import util

__all__ = (
    'parse_hierarchical_utilization_report',
    'HierarchicalUtilization',
)

_logger = logging.getLogger().getChild(__name__)


class HierarchicalUtilization:
  """Semantic-agnostic hierarchical utilization."""
  device: str
  parent: Optional['HierarchicalUtilization']
  children: List['HierarchicalUtilization']
  instance: str
  schema: Dict[str, int]
  items: Tuple[str, ...]

  def __init__(
      self,
      device: str,
      instance: str,
      schema: Dict[str, int],
      items: Tuple[str, ...],
      parent: Optional['HierarchicalUtilization'] = None,
  ) -> None:
    if len(schema) != len(items):
      raise TypeError('mismatching schema and items')
    self.device = device
    self.parent = parent
    self.children = []
    if parent is not None:
      parent.children.append(self)
    self.instance = instance
    self.schema = schema
    self.items = items

  def __getitem__(self, key: str) -> str:
    return self.items[self.schema[key]]

  def __str__(self) -> str:
    parent = None
    if self.parent is not None:
      parent = self.parent.instance
    return '\n'.join((
        '',
        f'instance: {self.instance}',
        f'parent: {parent}',
        *(f'{key}: {value}' for key, value in zip(self.schema, self.items)),
    ))


def parse_hierarchical_utilization_report(
    rpt_file: TextIO) -> HierarchicalUtilization:
  """Parse hierarchical utilization report.

  This is a compromise where Vivado won't export structured report from scripts.
 """

  class ParseState(enum.Enum):
    PROLOG = 0
    HEADER = 1
    BODY = 2
    EPILOG = 3

  parse_state = ParseState.PROLOG
  stack: List[HierarchicalUtilization] = []
  device = ''

  for line in rpt_file:
    line = line.strip()
    items = line.split()
    if len(items) == 4 and items[:3] == ['|', 'Device', ':']:
      device = items[3]
      continue
    if set(line) == {'+', '-'}:
      if parse_state == ParseState.PROLOG:
        parse_state = ParseState.HEADER
      elif parse_state == ParseState.HEADER:
        parse_state = ParseState.BODY
      elif parse_state == ParseState.BODY:
        parse_state = ParseState.EPILOG
      else:
        raise util.InputError('unexpected table separator line')
      continue

    if parse_state == ParseState.HEADER:
      instance, items = get_items(line)
      assert instance.lstrip() == 'Instance'
      schema = {x.lstrip(): i for i, x in enumerate(items)}

    elif parse_state == ParseState.BODY:
      instance, items = get_items(line)
      while (len(instance) - len(instance.lstrip(' '))) // 2 < len(stack):
        stack.pop()
      instance = instance.lstrip()
      parent = stack[-1] if stack else None
      stack.append(
          HierarchicalUtilization(device, instance, schema, items, parent))

  return stack[0]


def get_items(line: str) -> Tuple[str, Tuple[str, ...]]:
  """Split a table line into items.

  Args:
      line (str): A line in a report table.

  Returns:
      Tuple[str, ...]: Fields splitted by '|' with blank on the right stripped.
  """
  items = line.strip().strip('|').split('|')
  return items[0].rstrip(), tuple(x.strip() for x in items[1:])
